#include <unistd.h> /* for getopt */
#include <iostream>
#include <fstream>
#include <sstream>
#include <string.h>
#include <math.h>
#include <map>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

class MODEL
{  
  // for training
  int _N; // #users
  int _I; // #actions
  int _C; // #time-of-day windows
  int _K; // #mixtures  
  
  double _T; //observation period (minutes)
  double _prd; //minutes of a day
  int _D; //#days = _T / _prd
  
  int* _Mx;
  int** _ind;
  double** _ts;
  int** _tc;
  double* _th;
  int** _tl; // split index
  
  // for prediction
  int _Nt;
  int _It;
  int _Ct;  
  double _Tt;
  int _Dt;  
  int* _Mxt;
  int** _indt;
  double** _tst;
  int** _tct;
  double* _tht;
  int** _tlt;
  
  // model parameters  
  double* _beta; 
  double* _mu;
  double* _sigma;
  double* _theta;
  double* _phi;
  double* _omega;
  double* _gamma;
  double* _kappa;  
  double* _pref;
  
  double*** _trg0; // gaussian baserate
  double*** _trg1; // short-term
  double*** _trg2; // long-term
  double*** _trg3; // preference
  
  double logsumexp ( double x, double y, bool flg );
  double uniform ();
  double loglikelihood ( int& N, int& I, double& T, int& D, int& C, int*& Mx,  int**& ind, double**& ts, double*& th, int**& tl, int**& tc );
  
  void estep ();
  void updatebeta ();
  void updatemu ();
  void updatesigma ();
  void updatetheta ();
  void updatephi ();
  void updategamma ();
  void updateomega ();
  void updatekappa ();
  void updatepref ();
  void readtd ( char* fn, int& N, int& I, double& T, int& D, int& C, int*& Mx, int**& ind, double**& ts, double*& th, int**& tl, int**& tc );
  
public :
  MODEL ( char* fn, char* testfn, bool testflag, int K, double prd );
  void inference ( int maxstep );
  void outmodel ( char* );
  void readmodel ( char* );
  void initialize ( double initscale );
  void accuracy ();
};

double MODEL::uniform ( void ){ return ( ( double ) rand () + 1.0 ) / ( ( double ) RAND_MAX + 2.0 ); }

double MODEL::logsumexp ( double x, double y, bool flg )
{
  if ( flg ) return y; // init mode
  if ( x == y ) return x + 0.69314718055;  // log(2)
  double vmin = min ( x, y );
  double vmax = max ( x, y );
  if ( vmax > vmin + 50 ) {
    return vmax;
  } else {
    return vmax + log ( exp ( vmin - vmax ) + 1.0 );
  }
}

void MODEL::initialize ( double initscale )
{
  // gaussian baserate
  _beta = new double [ _I * _K ];
  _mu = new double [ _I * _K ];
  _sigma = new double [ _I * _K ];
  for ( int i = 0; i < _I * _K; i ++ ) _beta[i] = ( double ) random () / RAND_MAX;
  for ( int i = 0; i < _I; i ++ )
    for (int k = 0; k < _K; k ++){
      _mu[i*_K+k] = ( _prd / ( _K + 1 ) ) * ( k + 1 );
      _sigma[i*_K+k] = ( _prd / ( _K + 1 ) ) / 2;
    }
  // short-term
  _theta = new double [ _I * _I ];
  _omega = new double [ _I * _I ];
  for ( int i = 0; i < _I * _I; i ++ ){
    _theta[i] = ( double ) random () / RAND_MAX;
    _omega[i] = ( double ) random () / RAND_MAX;
  }
  // long-term
  _phi = new double [ _I * _C ];
  _gamma = new double [ _I * _C ];
  _kappa = new double [ _I * _C ];
  for ( int i = 0; i < _I * _C; i ++ ){
    _phi[i] = ( double ) random () / RAND_MAX;
    _kappa[i] = 5;
    _gamma[i] = 1 / pow( initscale, _kappa[i] ) * ( _kappa[i] - 1 ) / _kappa[i];
  }
  // preference
  _pref = new double [ _I * _N ];
  for ( int i = 0; i < _I * _N; i ++ ) _pref[i] = ( double ) random () / RAND_MAX;
  
  // trigger
  _trg0 = new double** [ _N ]; // gaussian baserate
  _trg1 = new double** [ _N ]; // short-term
  _trg2 = new double** [ _N ]; // long-term
  _trg3 = new double** [ _N ]; // preference
  for ( int u = 0; u < _N; u ++ ){
    _trg0[u] = new double* [ _Mx[u] ];
    _trg1[u] = new double* [ _Mx[u] ];
    _trg2[u] = new double* [ _Mx[u] ];
    _trg3[u] = new double* [ _Mx[u] ];
    for ( int n = 0; n < _Mx[u]; n ++ ){
      _trg0[u][n] = new double[ ( _tl[u][n] + 1 ) * _K ];
      _trg1[u][n] = new double [ n ];
      _trg2[u][n] = new double [ n ];
      _trg3[u][n] = new double [ 1 ];
    }
  }
}

void MODEL::inference ( int maxstep )
{
  cout << "#EM-steps\tLog-likelihood" << endl;
  double preL = 0.0;
  double mindiff = 1e-10;
  for ( int step = 0; step < maxstep; step ++ ) {    
    
    // E-step
    estep ();
    
    // M-step
    updatepref ();    
    updatetheta ();
    updateomega ();    
    
    updatekappa ();
    updatephi ();
    updategamma ();
    
    updatemu ();
    updatesigma ();
    updatebeta ();
    
    // log likelihood
    double L = loglikelihood ( _N, _I, _T, _D, _C, _Mx, _ind, _ts, _th, _tl, _tc );    
    double diff = L - preL;
    cout << step + 1 << "\t\t" << L << endl;
    if ( step > 0 && fabs ( diff ) < mindiff ) break;
    preL = L;
  }
}

void MODEL::estep ()
{
  for ( int u = 0; u < _N; u ++ ){
    for ( int n = 0; n < _Mx[u]; n ++ ){      
      
      double sum; int cnt = 0;      
      // gaussian baserate
      for ( int l = _tl[u][n]; l <= _tl[u][n]; l ++ )
        for ( int k = 0; k < _K; k ++ ){
          sum = logsumexp ( sum, log ( _beta[_ind[u][n]*_K+k] ) - 0.5 * log ( 2 * M_PI * pow ( _sigma[_ind[u][n]*_K+k], 2 ) ) - 0.5 * pow ( _sigma[_ind[u][n]*_K+k], - 2 ) * pow (  _ts[u][n] - _th[l] - _mu[_ind[u][n]*_K+k], 2 ), cnt == 0 ); cnt ++;
        }
      // short-term
      for ( int m = 0; m < n; m ++ ){
        sum = logsumexp ( sum, log ( _theta[_ind[u][m]*_I+_ind[u][n]] ) + log ( _omega[_ind[u][m]*_I+_ind[u][n]] ) - _omega[_ind[u][m]*_I+_ind[u][n]] * ( _ts[u][n] - _ts[u][m] ), ( cnt == 0 ) ); cnt ++;
      }
      // long-term
      for ( int m = 0; m < n; m ++ )
        if ( _ind[u][m] == _ind[u][n] ){
          sum = logsumexp ( sum, log ( _phi[_ind[u][m]*_C+_tc[u][m]] ) + log ( _gamma[_ind[u][m]*_C+_tc[u][m]] ) + log ( _kappa[_ind[u][m]*_C+_tc[u][m]] ) + ( _kappa[_ind[u][m]*_C+_tc[u][m]] - 1 ) * log ( _ts[u][n] - _ts[u][m] ) - _gamma[_ind[u][m]*_C+_tc[u][m]] * pow ( _ts[u][n] - _ts[u][m], _kappa[_ind[u][m]*_C+_tc[u][m]]), ( cnt == 0 ) ); cnt ++;
        }
      // preference
      if ( _pref[u*_I+_ind[u][n]] != 0 ){ sum = logsumexp ( sum, log ( _pref[u*_I+_ind[u][n]] ), ( cnt == 0 ) ); cnt ++; }
      if ( isnan ( sum ) || isinf ( sum ) ){ cerr << "Error at estep" << endl; exit ( 1 ); }
      
      // update trigger
      // trg0, gaussian baserate
      for ( int i = 0; i < _K * ( _tl[u][n] + 1 ); i ++ ) _trg0[u][n][i] = 0.0;
      for ( int l = _tl[u][n]; l <= _tl[u][n]; l ++ )
        for ( int k = 0; k < _K; k ++ )
          _trg0[u][n][l*_K+k] = log ( _beta[_ind[u][n]*_K+k] ) - 0.5 * log ( 2 * M_PI * pow ( _sigma[_ind[u][n]*_K+k], 2 ) ) - 0.5 * pow ( _sigma[_ind[u][n]*_K+k], - 2 ) * pow ( _ts[u][n] - _th[l] - _mu[_ind[u][n]*_K+k], 2 ) - sum;
      // trg1, short-term
      for ( int i = 0; i < n; i ++ ) _trg1[u][n][i] = 0.0;
      for ( int m = 0; m < n; m ++ )
	_trg1[u][n][m] = log ( _theta[_ind[u][m]*_I+_ind[u][n]] ) + log ( _omega[_ind[u][m]*_I+_ind[u][n]] ) - _omega[_ind[u][m]*_I+_ind[u][n]] * ( _ts[u][n] - _ts[u][m] ) - sum;      
      // trg2, long-term
      for ( int i = 0; i < n; i ++ ) _trg2[u][n][i] = log ( 0 );
      for ( int m = 0; m < n; m ++ )
        if ( _ind[u][m] == _ind[u][n] )
          _trg2[u][n][m] = log ( _phi[_ind[u][m]*_C+_tc[u][m]] ) + log ( _gamma[_ind[u][m]*_C+_tc[u][m]] ) + log ( _kappa[_ind[u][m]*_C+_tc[u][m]] ) + ( _kappa[_ind[u][m]*_C+_tc[u][m]] - 1 ) * log ( _ts[u][n] - _ts[u][m] ) - _gamma[_ind[u][m]*_C+_tc[u][m]] * pow ( _ts[u][n] - _ts[u][m], _kappa[_ind[u][m]*_C+_tc[u][m]]) - sum;
      // trg3, preference
      _trg3[u][n][0] = log ( _pref[u*_I+_ind[u][n]] ) - sum;
    }
  }
}

void MODEL::updatepref ()
{
  double* sum = new double [ _I ];
  for ( int u = 0; u < _N; u ++ ){
    for ( int i = 0; i < _I; i ++ ) sum[i] = 0.0;
    for ( int n = 0; n < _Mx[u]; n ++ )
      if ( isnan ( _trg3[u][n][0] ) || isinf ( _trg3[u][n][0] ) ){
	sum[_ind[u][n]] += 0.0;
      }else{
	sum[_ind[u][n]] += exp ( _trg3[u][n][0] );
      }
    for ( int i = 0; i < _I; i ++ ){
      _pref[u*_I+i] = sum[i] / _T; // update
      if ( ( isnan ( _pref[u*_I+i] ) ) || ( isinf ( _pref[u*_I+i] ) ) ){ _pref[u*_I+i] = 0.0; }
    }
  }
  delete [] sum;
}

void MODEL::updatebeta ()
{
  double* sum = new double [ _I * _K ]; for ( int i = 0; i < _I * _K; i ++ ) sum[i] = 0.0;  
  for ( int z = 0; z < _K; z ++)
    for ( int u = 0; u < _N; u ++ )
      for ( int n = 0; n < _Mx[u]; n ++ )
        for ( int l = _tl[u][n]; l <= _tl[u][n]; l ++ )
          sum[_ind[u][n]*_K+z] += exp ( _trg0[u][n][l*_K+z] );  
  for ( int z = 0; z < _K; z ++ )
    for (int x = 0; x < _I; x ++){
      double current = _beta[x*_K+z];
      _beta[x*_K+z] = sum[x*_K+z] / ( _N * _D * 0.5 * ( erf ( ( _prd - _mu[x*_K+z] ) / ( sqrt ( 2 ) * _sigma[x*_K+z] ) ) + erf ( _mu[x*_K+z] / ( sqrt ( 2 ) * _sigma[x*_K+z] ) ) ) ); // update
      if ( isnan ( _beta[x*_K+z] ) || isinf ( _beta[x*_K+z] ) ){ _beta[x*_K+z] = current; }
    }
  delete [] sum;
}

void MODEL::updatemu ()
{
  double* epsilon = new double [ _I * _K ];
  for ( int i = 0; i < _I * _K; i ++ ){ epsilon[i] = log ( 1 / _mu[i] ); } // mu -> epsilon
  
  int maxstep = 5;
  double mindiff = 1e-10;
  double sum1, sum2;
  
  for (int x = 0; x < _I; x ++){
    for ( int z = 0; z < _K; z ++ ){
      double current = epsilon[x*_K+z];
      
      // Newton method
      for ( int step = 0; step < maxstep; step ++ ) {
        double previous = epsilon[x*_K+z];
	sum1 = 0.0; sum2 = 0.0;
	
        for ( int u = 0; u < _N; u ++ )
          for ( int n = 0; n < _Mx[u]; n ++ )
            if ( _ind[u][n] == x ){
	      sum1 += exp ( _trg0[u][n][_tl[u][n]*_K+z] ) * ( 1 / pow ( _sigma[x*_K+z], 2 ) ) * exp ( - 2 * epsilon[x*_K+z] ) * ( 1 - ( _ts[u][n] - _th[_tl[u][n]] ) * exp ( epsilon[x*_K+z] ) );
	      sum2 += exp ( _trg0[u][n][_tl[u][n]*_K+z] ) * ( 1 / pow ( _sigma[x*_K+z], 2 ) ) * exp ( - 2 * epsilon[x*_K+z] ) * ( ( _ts[u][n] - _th[_tl[u][n]] ) * exp ( epsilon[x*_K+z] ) - 2 );
	    }
        sum1 += - 0.5 * _N * _D * _beta[x*_K+z] * sqrt ( 2 / M_PI ) * ( 1 / _sigma[x*_K+z] ) * exp ( - epsilon[x*_K+z] )
	  * ( exp ( - 0.5 * pow ( _sigma[x*_K+z], - 2 ) * pow ( exp ( - epsilon[x*_K+z] ) - _prd, 2 ) ) - exp ( - 0.5 * pow ( _sigma[x*_K+z], - 2 ) * exp ( - 2 * epsilon[x*_K+z] ) ) );
	sum2 += - 0.5 * _N * _D * _beta[x*_K+z] * sqrt ( 2 / M_PI ) * ( 1 / _sigma[x*_K+z] )
	  * ( exp ( - 0.5 * pow ( _sigma[x*_K+z], - 2 ) * pow ( _prd - exp ( - epsilon[x*_K+z] ), 2 ) - epsilon[x*_K+z] ) * ( - pow ( _sigma[x*_K+z], - 2 ) * exp ( - epsilon[x*_K+z] ) * ( _prd - exp ( - epsilon[x*_K+z] ) ) - 1 )
	      - exp ( - 0.5 * pow ( _sigma[x*_K+z], - 2 ) * exp ( - 2 * epsilon[x*_K+z] ) - epsilon[x*_K+z] ) * ( pow ( _sigma[x*_K+z], - 2 ) * exp ( - 2 * epsilon[x*_K+z] ) - 1 ) );
	
        epsilon[x*_K+z] -= sum1 / sum2; // update
	if ( ( ( 1 / exp ( epsilon[x*_K+z] ) ) >= _prd ) || ( isnan ( 1 / exp ( epsilon[x*_K+z] ) ) ) || ( isinf ( 1 / exp ( epsilon[x*_K+z] ) ) ) ){ epsilon[x*_K+z] = current; break; }
	double diff = epsilon[x*_K+z] - previous; if ( step > 0 && fabs ( diff ) < mindiff ) break;
      }
    }
  }
  for ( int i = 0; i < _I * _K; i ++ ) _mu[i] = 1 / exp ( epsilon[i] ); // epsilon -> mu
  delete [] epsilon;
}

void MODEL::updatesigma ()
{
  double* zeta = new double [ _I * _K ];
  for ( int i = 0; i < _I * _K; i ++ ){ zeta[i] = log ( 1 / pow ( _sigma[i], 2 ) ); } // sigma -> zeta
  
  int maxstep = 5;
  double mindiff = 1e-10;
  double sum1, sum2, term;
  
  for (int x = 0; x < _I; x ++){
    for ( int z = 0; z < _K; z ++ ){
      double current = zeta[x*_K+z];
      
      // Newton method
      for ( int step = 0; step < maxstep; step ++ ) {
	double previous = zeta[x*_K+z];
	sum1 = 0.0; sum2 = 0.0;
	
	for ( int u = 0; u < _N; u ++ )
	  for ( int n = 0; n < _Mx[u]; n ++ )
	    if ( _ind[u][n] == x ){
	      sum1 += 0.5 * exp ( _trg0[u][n][_tl[u][n]*_K+z] ) * ( 1 - exp ( zeta[x*_K+z] ) * pow ( _ts[u][n] - _th[_tl[u][n]] - _mu[x*_K+z], 2 ) );
	      sum2 += - 0.5 * exp ( _trg0[u][n][_tl[u][n]*_K+z] ) * exp ( zeta[x*_K+z] ) * pow ( _ts[u][n] - _th[_tl[u][n]] - _mu[x*_K+z], 2 );
	    }
	
	term = pow( exp ( zeta[x*_K+z] ), 0.5 ) * _mu[x*_K+z] * exp( - 0.5 * exp ( zeta[x*_K+z] ) * pow ( _mu[x*_K+z], 2 ) );
	term += pow( exp ( zeta[x*_K+z] ), 0.5 ) * ( _prd - _mu[x*_K+z] ) * exp( - 0.5 * exp ( zeta[x*_K+z] ) * pow ( _prd - _mu[x*_K+z], 2 ) );
	sum1 += - 0.5 * _N * _D * _beta[x*_K+z] * ( 1 / sqrt ( 2 * M_PI ) ) * term;
	
	term = ( _prd - _mu[x*_K+z] ) * exp( - 0.5 * exp( zeta[x*_K+z] ) * pow( _prd - _mu[x*_K+z], 2 ) );
        term += - pow( _prd - _mu[x*_K+z], 3 ) * exp( zeta[x*_K+z] - 0.5 * exp( zeta[x*_K+z] ) * pow( _prd - _mu[x*_K+z], 2 ) );
        term += _mu[x*_K+z] * exp( - 0.5 * exp( zeta[x*_K+z] ) * pow( _mu[x*_K+z], 2 ) );
        term += - pow( _mu[x*_K+z], 3 ) * exp( zeta[x*_K+z] - 0.5 * exp( zeta[x*_K+z] ) * pow( _mu[x*_K+z], 2 ) );
        sum2 += - 0.5 * _N * _D * _beta[x*_K+z] * ( 1 / sqrt ( 2 * M_PI ) ) * 0.5 * pow( exp ( zeta[x*_K+z] ), 0.5 ) * term;
	
	zeta[x*_K+z] -= sum1 / sum2; // update
	
	if ( isnan ( sqrt ( 1 / exp ( zeta[x*_K+z] ) ) ) || isinf ( sqrt ( 1 / exp ( zeta[x*_K+z] ) ) ) ){ zeta[x*_K+z] = current; break; }
	if ( fabs ( sqrt ( 1 / exp ( zeta[x*_K+z] ) ) - sqrt ( 1 / exp ( previous ) ) ) > 300 ){ zeta[x*_K+z] = previous; break; }
	double diff = zeta[x*_K+z] - previous; if ( step > 0 && fabs ( diff ) < mindiff ) break;
      }
    }
  }
  for ( int i = 0; i < _I * _K; i ++ ){ _sigma[i] = sqrt ( 1 / exp ( zeta[i] ) ); } // zeta -> sigma
  delete [] zeta;
}

void MODEL::updatetheta ()
{
  double* sum1 = new double [ _I * _I ];
  double* sum2 = new double [ _I * _I ];
  double* iunx = new double [ _I ];  
  for ( int i = 0; i < _I * _I; i ++ ){ sum1[i] = log ( 0 ); sum2[i] = log ( 0 ); }
  for ( int i = 0; i < _I; i ++ ){ iunx[i] = 0.0; }
  
  for ( int u = 0; u < _N; u ++ )
    for ( int n = 0; n < _Mx[u]; n ++ ){
      iunx[_ind[u][n]] ++;
      for ( int y = 0; y < _I; y ++ )
        sum1[_ind[u][n]*_I+y] = logsumexp ( sum1[_ind[u][n]*_I+y], - _omega[_ind[u][n]*_I+y] * ( _T - _ts[u][n]), isinf ( sum1[_ind[u][n]*_I+y] ) );      
      for( int m = 0; m < n; m ++ )
        sum2[_ind[u][m]*_I+_ind[u][n]] = logsumexp ( sum2[_ind[u][m]*_I+_ind[u][n]], _trg1[u][n][m], isinf ( sum2[_ind[u][m]*_I+_ind[u][n]] ) );      
    }
  for ( int x = 0; x < _I; x ++ )
    for ( int y = 0; y < _I; y ++ ){
      _theta[x*_I+y] = exp ( sum2[x*_I+y] ) / ( iunx[x] - exp ( sum1[x*_I+y] ) ); // update
      if ( ( isnan ( _theta[x*_I+y] ) ) || ( isinf ( _theta[x*_I+y] ) ) ){ _theta[x*_I+y] = 0.0; }
    }
  delete [] sum1;
  delete [] sum2;
  delete [] iunx;
}

void MODEL::updateomega ()
{
  double* sum1 = new double [ _I * _I ];
  double* sum2 = new double [ _I * _I ];
  double* sum3 = new double [ _I * _I ];
  
  for ( int i = 0; i < _I * _I; i ++ ){ sum1[i] = log ( 0 ); sum2[i] = log ( 0 ); sum3[i] = log ( 0 ); }
  for ( int u = 0; u < _N; u ++ )
    for ( int n = 0; n < _Mx[u]; n ++ ){
      for ( int y = 0; y < _I; y ++ )
        if ( _theta[_ind[u][n]*_I+y] != 0 )
	  sum2[_ind[u][n]*_I+y] = logsumexp ( sum2[_ind[u][n]*_I+y], log ( _theta[_ind[u][n]*_I+y] ) + log ( _T - _ts[u][n] ) - _omega[_ind[u][n]*_I+y] * ( _T - _ts[u][n] ), isinf ( sum2[_ind[u][n]*_I+y] ) );
      for( int m = 0; m < n; m ++ ){
        sum1[_ind[u][m]*_I+_ind[u][n]] = logsumexp ( sum1[_ind[u][m]*_I+_ind[u][n]], _trg1[u][n][m] + log ( _ts[u][n] - _ts[u][m] ), isinf ( sum1[_ind[u][m]*_I+_ind[u][n]] ) );
        sum3[_ind[u][m]*_I+_ind[u][n]] = logsumexp ( sum3[_ind[u][m]*_I+_ind[u][n]], _trg1[u][n][m], isinf ( sum3[_ind[u][m]*_I+_ind[u][n]] ) );
      }
    }
  for ( int x = 0; x < _I; x ++ )
    for ( int y = 0; y < _I; y ++ ){
      double current = _omega[x*_I+y];
      _omega[x*_I+y] = exp ( sum3[x*_I+y] ) / ( ( exp ( sum1[x*_I+y] ) ) + ( exp ( sum2[x*_I+y] ) ) ); // update
      if ( isnan ( _omega[x*_I+y] ) || isinf ( _omega[x*_I+y] ) ){ _omega[x*_I+y] = current; }
    }
  delete [] sum1;
  delete [] sum2;
  delete [] sum3;
}

void MODEL::updatephi ()
{
  double* sum1 = new double [ _I * _C ];
  double* sum2 = new double [ _I * _C ];
  double* iunx = new double [ _I * _C ];
  
  for ( int i = 0; i < _I * _C; i ++ ){ sum1[i] = log ( 0 ); sum2[i] = log ( 0 ); iunx[i] = 0.0; }
  for ( int u = 0; u < _N; u ++ )
    for ( int n = 0; n < _Mx[u]; n ++ ){
      iunx[_ind[u][n]*_C+_tc[u][n]] ++;
      sum1[_ind[u][n]*_C+_tc[u][n]] = logsumexp ( sum1[_ind[u][n]*_C+_tc[u][n]], - _gamma[_ind[u][n]*_C+_tc[u][n]] * pow ( _T - _ts[u][n], _kappa[_ind[u][n]*_C+_tc[u][n]] ), isinf ( sum1[_ind[u][n]*_C+_tc[u][n]] ) );
      for( int m = 0; m < n; m ++ )
        if ( _ind[u][m] == _ind[u][n] )
          sum2[_ind[u][m]*_C+_tc[u][m]] = logsumexp ( sum2[_ind[u][m]*_C+_tc[u][m]], _trg2[u][n][m], isinf ( sum2[_ind[u][m]*_C+_tc[u][m]] ) );
    }
  for ( int x = 0; x < _I; x ++ )
    for ( int y = 0; y < _C; y ++ ){
      _phi[x*_C+y] = exp ( sum2[x*_C+y] ) / ( iunx[x*_C+y] - exp ( sum1[x*_C+y] ) ); // update
      if ( ( isnan ( _phi[x*_C+y] ) ) || ( isinf ( _phi[x*_C+y] ) ) ){ _phi[x*_C+y] = 0.0; }
    }
  delete [] sum1;
  delete [] sum2;
  delete [] iunx;
}

void MODEL::updategamma ()
{
  double* sum1 = new double [ _I * _C ];
  double* sum2 = new double [ _I * _C ];
  double* sum3 = new double [ _I * _C ];
  for ( int i = 0; i < _I * _C; i ++ ){ sum1[i] = log ( 0 ); sum2[i] = log ( 0 ); sum3[i] = log ( 0 ); }
  for ( int u = 0; u < _N; u ++ )
    for ( int n = 0; n < _Mx[u]; n ++ ){
      for( int m = 0; m < n; m ++ )
        if ( _ind[u][m] == _ind[u][n] ){
          sum3[_ind[u][n]*_C+_tc[u][m]] = logsumexp ( sum3[_ind[u][n]*_C+_tc[u][m]], _trg2[u][n][m], isinf ( sum3[_ind[u][n]*_C+_tc[u][m]] ) );
          sum1[_ind[u][n]*_C+_tc[u][m]] = logsumexp ( sum1[_ind[u][n]*_C+_tc[u][m]], _trg2[u][n][m] + _kappa[_ind[u][n]*_C+_tc[u][m]] * log ( _ts[u][n] - _ts[u][m] ), isinf ( sum1[_ind[u][n]*_C+_tc[u][m]] ) );
        }
      if ( _phi[_ind[u][n]*_C+_tc[u][n]] != 0 )
        sum2[_ind[u][n]*_C+_tc[u][n]] = logsumexp ( sum2[_ind[u][n]*_C+_tc[u][n]], log ( _phi[_ind[u][n]*_C+_tc[u][n]] ) + _kappa[_ind[u][n]*_C+_tc[u][n]] * log ( _T - _ts[u][n] ) - _gamma[_ind[u][n]*_C+_tc[u][n]] * pow ( _T - _ts[u][n], _kappa[_ind[u][n]*_C+_tc[u][n]] ), isinf ( sum2[_ind[u][n]*_C+_tc[u][n]] ) );      
    }
  for ( int x = 0; x < _I; x ++ )
    for ( int y = 0; y < _C; y ++ ){
      double current = _gamma[x*_C+y];
      _gamma[x*_C+y] = exp ( sum3[x*_C+y] ) / ( exp ( sum1[x*_C+y] ) + exp ( sum2[x*_C+y] ) ); // update
      if ( ( isnan ( _gamma[x*_C+y] ) ) || ( isinf ( _gamma[x*_C+y] ) ) ){ _gamma[x*_C+y] = current; }
    }
  delete [] sum1;
  delete [] sum2;
  delete [] sum3;
}

void MODEL::updatekappa ()
{
  int maxstep = 50;
  double mindiff = 1e-10;
  double sum1, sum2, sum3, sum4, sum5;
  int cnt;
  
  for ( int x = 0; x < _I; x ++ ){
    for( int y = 0; y < _C; y ++ ){
      double current = _kappa[x*_C+y];     
      
      // Newton method
      for ( int step = 0; step < maxstep; step ++ ) {
        sum1 = 0.0; sum2 = 0.0; sum3 = 0.0; sum4 = 0.0; sum5 = 0.0; cnt = 0;
	double previous = _kappa[x*_C+y];
        for ( int u = 0; u < _N; u ++ ){
          for ( int n = 0; n < _Mx[u]; n ++ ){
            for( int m = 0; m < n; m ++ ){
	      if ( ( _ind[u][m] == _ind[u][n] ) && ( _ind[u][n] == x ) && ( _tc[u][m] == y ) ){
		sum1 += exp ( _trg2[u][n][m] ) * ( 1 / _kappa[x*_C+y] + log ( _ts[u][n] - _ts[u][m] ) - _gamma[x*_C+y] * pow ( _ts[u][n] - _ts[u][m], _kappa[x*_C+y] ) * log ( _ts[u][n] - _ts[u][m] ) );
		sum2 += exp ( _trg2[u][n][m] ) * ( - _gamma[x*_C+y] * pow ( _ts[u][n] - _ts[u][m], _kappa[x*_C+y] ) * pow ( log ( _ts[u][n] - _ts[u][m] ), 2 ) - 1 / pow ( _kappa[x*_C+y], 2 ) );
              }
            }
            if ( ( _ind[u][n] == x ) && ( _tc[u][n] == y ) ){
              sum3 = logsumexp ( sum3, log ( _phi[x*_C+y] ) + log ( _gamma[x*_C+y] ) + _kappa[x*_C+y] * log ( _T - _ts[u][n] ) + log ( log ( _T - _ts[u][n] ) ) - _gamma[x*_C+y] * pow ( _T - _ts[u][n], _kappa[x*_C+y] ), cnt == 0 );
              sum4 = logsumexp ( sum4, log ( _phi[x*_C+y] ) + log ( _gamma[x*_C+y] ) + _kappa[x*_C+y] * log ( _T - _ts[u][n] ) + log ( pow ( log ( _T - _ts[u][n] ), 2 ) ) - _gamma[x*_C+y] * pow ( _T - _ts[u][n], _kappa[x*_C+y] ) + log ( _gamma[x*_C+y] * pow ( _T - _ts[u][n], _kappa[x*_C+y] ) ), cnt == 0 );
	      sum5 = logsumexp ( sum5, log ( _phi[x*_C+y] ) + log ( _gamma[x*_C+y] ) + _kappa[x*_C+y] * log ( _T - _ts[u][n] ) + log ( pow ( log ( _T - _ts[u][n] ), 2 ) ) - _gamma[x*_C+y] * pow ( _T - _ts[u][n], _kappa[x*_C+y] ), cnt == 0);
              cnt ++;
            }
          }
        }
	_kappa[x*_C+y] -= ( sum1 - exp ( sum3 ) ) / ( sum2 + exp ( sum4 ) - exp( sum5 ) ); // update
	if ( ( isnan ( _kappa[x*_C+y] ) ) || ( isinf ( _kappa[x*_C+y] ) ) || ( _kappa[x*_C+y] <= 1 ) ){ _kappa[x*_C+y] = current; break; }
        double diff = _kappa[x*_C+y] - previous; if ( step > 0 && fabs ( diff ) < mindiff ) break;
      }
    }
  }
}

double MODEL::loglikelihood ( int& N, int& I, double& T, int& D, int& C, int*& Mx,  int**& ind, double**& ts, double*& th, int**& tl, int**& tc )
{
  double LL = 0.0;
  double sum = 0.0;
  
  for ( int u = 0; u < N; u ++ )
    for ( int n = 0; n < Mx[u]; n ++ ){      
      sum = 0.0;
      
      // preference
      sum += _pref[u*I+ind[u][n]];
      // gaussian baserate
      for ( int k = 0; k < _K; k ++ )
	if ( ! ( isnan ( pow ( _sigma[ind[u][n]*_K+k], 2 ) ) || ( isinf ( pow ( _sigma[ind[u][n]*_K+k], 2 ) ) ) || ( pow ( _sigma[ind[u][n]*_K+k], 2 ) == 0 ) ) )
	  sum += _beta[ind[u][n]*_K+k] / sqrt( 2 * M_PI * pow ( _sigma[ind[u][n]*_K+k], 2 ) ) * exp( - ( pow( ts[u][n] - th[tl[u][n]] - _mu[ind[u][n]*_K+k], 2 ) / ( 2 * pow ( _sigma[ind[u][n]*_K+k], 2 ) ) ) );
      // short-term
      for ( int m = 0; m < n; m ++ )
	sum += _theta[ind[u][m]*I+ind[u][n]] * _omega[ind[u][m]*I+ind[u][n]] * exp( - _omega[ind[u][m]*I+ind[u][n]] * ( ts[u][n] - ts[u][m] ) );
      // long-term
      for ( int m = 0; m < n; m ++ ) 
	if ( ind[u][m] == ind[u][n] )
	  sum += _phi[ind[u][m]*C+tc[u][m]] * _gamma[ind[u][m]*C+tc[u][m]] * _kappa[ind[u][m]*C+tc[u][m]]
	    * pow( ts[u][n] - ts[u][m], _kappa[ind[u][m]*C+tc[u][m]] - 1 )
	    * exp( - _gamma[ind[u][m]*C+tc[u][m]] * pow( ts[u][n] - ts[u][m], _kappa[ind[u][m]*C+tc[u][m]] ) );
      if ( sum != 0 ) LL += log( sum );
    }
  
  // preference
  for ( int i = 0; i < I * N; i ++ ) LL -= T * _pref[i];  
  // gaussian baserate
  sum = 0.0;
  for ( int i = 0; i < I; i ++ )
    for ( int k = 0; k < _K; k ++ )
      sum += ( _beta[i*_K+k] / 2 ) * ( erf ( ( _prd - _mu[i*_K+k] ) / ( sqrt ( 2 ) * _sigma[i*_K+k] ) ) + erf ( _mu[i*_K+k] / ( sqrt ( 2 ) * _sigma[i*_K+k] ) ) );
  LL -= N * D * sum;
  // short-term
  sum = 0.0;
  for ( int u = 0; u < N; u ++ )
    for ( int i = 0; i < I; i ++ )
      for ( int n = 0; n < Mx[u]; n ++ )
        sum += _theta[ind[u][n]*I+i] * ( 1 - exp ( - _omega[ind[u][n]*I+i] * ( T - ts[u][n] ) ) );
  /// long-term
  for ( int u = 0; u < N; u ++ )
    for ( int n = 0; n < Mx[u]; n ++ )
      sum += _phi[ind[u][n]*C+tc[u][n]] * ( 1 - exp ( - _gamma[ind[u][n]*C+tc[u][n]] * ( pow ( T - ts[u][n], _kappa[ind[u][n]*C+tc[u][n]] ) ) ) );
  LL -= sum;
  
  return LL;
}

void MODEL::readtd ( char* fn, int& N, int& I, double& T, int& D, int& C, int*& Mx, int**& ind, double**& ts, double*& th, int**& tl, int**& tc )
{
  string line; ifstream ifs( fn );
  if ( !ifs.is_open() ) { cerr << "Cannot open " << fn << endl; exit ( 1 ); }
  getline( ifs, line ); istringstream istrs( (char*)line.c_str() );
  istrs >> N; istrs >> I; istrs >> T; istrs >> C;
  Mx = new int [ N ];
  ind = new int* [ N ];
  ts = new double* [ N ];
  tc = new int* [ N ];
  int j = 0;
  string stmp;
  while ( getline( ifs, line ) ) {
    istringstream istrs( (char*)line.c_str() );
    istrs >> Mx[j];
    ind[j] = new int [Mx[j]];
    ts[j] = new double [Mx[j]];
    tc[j] = new int [Mx[j]];
    for ( int i = 0; i < Mx[j]; i ++ ) {
      istrs >> stmp;
      sscanf( (char*)stmp.c_str(), "%d:%lf:%d", &ind[j][i], &ts[j][i], &tc[j][i] );     
    }
    j ++;
  }
  D = ( int ) T / _prd;
  th = new double [ D ]; for ( int i = 0; i < D; i++ ){ th[i] = i * _prd; }
  tl = new int* [ N ];
  for ( int u = 0; u < N; u ++ ){
    tl[u] = new int [ Mx[u] ];
    for ( int n = 0; n < Mx[u]; n ++ ){
      int cnt = 0; for ( int l = 1; l < D; l ++ ){ if ( th[l] < ts[u][n] ){ cnt ++; }else{ break; } }
      tl[u][n] = cnt;
    }
  }
}

void MODEL::outmodel ( char* fn )
{
  ofstream ofs ( fn );
  if ( ofs ){    
    for ( int x = 0; x < _I; x ++ )
      for ( int y = 0; y < _K; y ++)
	ofs << x << " " << y << " " << _beta[x*_K+y] << " " << _mu[x*_K+y] << " " << _sigma[x*_K+y] << endl;
    for ( int x = 0; x < _I; x ++ )
      for ( int y = 0; y < _I; y ++ )
	ofs << x << " " << y << " " << _theta[x*_I+y] << " " << _omega[x*_I+y] << endl;
    for ( int x = 0; x < _I; x ++ )
      for ( int y = 0; y < _C; y ++ )
	ofs << x << " " << y << " " << _phi[x*_C+y] << " " << _gamma[x*_C+y] << " " << _kappa[x*_C+y] << endl;      
    for ( int u = 0; u < _N; u ++ )
      for ( int x = 0; x < _I; x ++ )
	ofs << u << " " << x << " " << _pref[u*_I+x] << endl;
  }
}

void MODEL::readmodel ( char* fn )
{
  _beta = new double [ _I * _K ];
  _mu = new double [ _I * _K ];
  _sigma = new double [ _I * _K ];
  _theta = new double [ _I * _I ];
  _omega = new double [ _I * _I ];
  _phi = new double [ _I * _C ];
  _gamma = new double [ _I * _C ];
  _kappa = new double [ _I * _C ];
  _pref = new double [ _I * _N ];
  
  string line; ifstream ifs( fn );
  if ( !ifs.is_open() ) { cerr << fn << ": cannot open" << endl; exit ( 1 ); }
  int i = 0; int x; int y; int u;
  while ( getline( ifs, line ) ) {
    istringstream istrs( (char*)line.c_str() );
    if ( i < _I * _K ){
      istrs >> x; istrs >> y; istrs >> _beta[x*_K+y]; istrs >> _mu[x*_K+y]; istrs >> _sigma[x*_K+y];
    }else if ( i < ( _I * _K + _I * _I ) ){
      istrs >> x; istrs >> y; istrs >> _theta[x*_I+y]; istrs >> _omega[x*_I+y];
    }else if ( i < ( _I * _K + _I * _I + _I * _C ) ){
      istrs >> x; istrs >> y; istrs >> _phi[x*_C+y]; istrs >> _gamma[x*_C+y]; istrs >> _kappa[x*_C+y];
    }else{
      istrs >> u; istrs >> x; istrs >> _pref[u*_I+x];
    }
    i ++;    
  }
}

void MODEL::accuracy ()
{ 
  double acc = 0.0;
  int testNM = 0;
  double* acc_item = new double [ _It ];
  int* testNM_item = new int [ _It ];
  double* rate = new double [ _It ];
  int topi = 1;
  
  for ( int x = 0; x < _It; x ++ ){ acc_item[x] = 0; testNM_item[x] = 0; }
  for ( int i = 0; i < _It; i++ ){ rate[i] = 0.0; }
  
  for ( int u = 0; u < _Nt; u ++ ){    
    for ( int x = 1; x < _Mxt[u]; x ++ ){
      int testaction = _indt[u][x];
      double testtime = _tst[u][x];
      
      map< double,int,greater<double> > results;
      for ( int i = 0; i < _It; i++ ){
	
	int cnt = 0;	
	/// preference
	rate[i] = logsumexp ( rate[i], log ( _pref[u*_I+i] ), ( cnt == 0 ) );
	cnt ++;
	/// gaussian baserate
	int lastl = 0; for ( int l = 0; l < _Dt; l ++ ){ if ( _tht[l] < testtime ){ lastl = l; } }
	for ( int k = 0; k < _K; k ++ ){
	  rate[i] = logsumexp ( rate[i], log ( _beta[i*_K+k] ) - 0.5 * log ( 2 * M_PI * pow ( _sigma[i*_K+k], 2 ) ) - 0.5 * ( 1 / pow ( _sigma[i*_K+k], 2 ) ) * pow ( testtime - _tht[lastl] - _mu[i*_K+k], 2 ), ( cnt == 0 ) );
	  cnt ++;
	}
	/// short-term
	for ( int n = 0; n < x; n ++ ){
	  rate[i] = logsumexp ( rate[i], log ( _theta[_indt[u][n]*_It+i] ) + log ( _omega[_indt[u][n]*_It+i] ) - _omega[_indt[u][n]*_It+i] * ( testtime - _tst[u][n] ), ( cnt == 0 ) );
	  cnt ++;
	}
	/// long-term
	for ( int n = 0; n < x; n ++ ){
	  if ( _indt[u][n] == i ){
	    rate[i] = logsumexp ( rate[i], log ( _phi[_indt[u][n]*_Ct+_tct[u][n]] ) + log ( _gamma[_indt[u][n]*_Ct+_tct[u][n]] ) + log ( _kappa[_indt[u][n]*_Ct+_tct[u][n]] ) + ( _kappa[_indt[u][n]*_Ct+_tct[u][n]] - 1 ) * log ( testtime - _tst[u][n] ) - _gamma[_indt[u][n]*_Ct+_tct[u][n]] * pow ( testtime - _tst[u][n], _kappa[_indt[u][n]*_Ct+_tct[u][n]]), ( cnt == 0 ) );
	    cnt ++;
	  }
	}
	results.insert ( map<double,int>::value_type( exp (rate[i]) , i) );
      }
      map <double, int>::iterator it = results.begin();
      int i = 0;
      while( it != results.end() ){
	if ( testaction == (*it).second ){
	  acc++;
	  acc_item[testaction]++;
	  break;
	}
	++it;
	i++;
	if ( i == topi ) break;
      }
      results.clear ();
      testNM ++;
      testNM_item[testaction]++;
    }
  }
  cout << "Accuracy: " << acc / testNM << endl;
  double mar = 0; for ( int x = 0; x < _It; x ++ ) mar += acc_item[x] / testNM_item[x]; mar /= _It;
  cout << "Macro-averaged Recall: " << mar << endl;
  
  delete [] acc_item;
  delete [] testNM_item;
  delete [] rate;
}

MODEL::MODEL ( char* fn, char* testfn, bool testflag, int K, double prd )
{
  _K = K; _prd = prd;
  readtd ( fn, _N, _I, _T, _D, _C, _Mx, _ind, _ts, _th, _tl, _tc );
  if ( testflag ){ readtd ( testfn, _Nt, _It, _Tt, _Dt, _Ct, _Mxt, _indt, _tst, _tht, _tlt, _tct ); }  
}

void usage ()
{
  fprintf( stderr, "Usage: tipas ifn ofn [-t testfile -m maxstep -z mixtures ]\n" );
  exit ( 1 );
}

int main ( int argc, char** argv )
{
  int ch;
  extern char *optarg;
  extern int optind, opterr;
  
  char* testfn;
  bool testflag = false;
  
  double prd = 60 * 24; // minutes of a day  
  int maxstep = 100; // # em steps
  int Z = 2; // # mixtures
  double initscale = 0.25 * prd; // 6 hours
  
  while ( ( ch = getopt ( argc, argv, "t:m:z:s:" ) ) != - 1 ){
    switch ( ch ){
    case 't':
      testflag = true;
      testfn = optarg;
      break;
    case 'm':
      maxstep = atoi( optarg );
      break;
    case 'z':
      Z = atoi( optarg );
      break;
    case 's':
      initscale = atof( optarg );
      break;
    default:
      usage ();
    }
  }
  
  argc -= optind;
  argv += optind;
  if ( argc != 2 ) usage ();
  
  char* ifn = argv[0];
  char* ofn = argv[1];  
  MODEL model = MODEL ( ifn, testfn, testflag, Z, prd );
  
  model.initialize ( initscale );
  model.inference ( maxstep );
  model.outmodel ( ofn );
  
  if ( testflag ){
    model.readmodel ( ofn );
    model.accuracy ();
  }
  cout << "Done." << endl;
}
